package de.uniwue.web.facade;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.UUID;
import java.util.function.Function;

import org.locationtech.jts.geom.*;
import org.locationtech.jts.geom.impl.CoordinateArraySequence;
import org.locationtech.jts.simplify.TopologyPreservingSimplifier;
import org.opencv.core.Mat;
import org.opencv.core.MatOfPoint;
import org.opencv.core.MatOfPoint2f;
import org.opencv.core.RotatedRect;
import org.opencv.core.Size;
import org.opencv.imgproc.Imgproc;

import de.uniwue.algorithm.data.MemoryCleaner;
import de.uniwue.algorithm.geometry.regions.RegionSegment;
import de.uniwue.algorithm.geometry.regions.type.RegionSubType;
import de.uniwue.algorithm.operators.Contourextractor;
import de.uniwue.algorithm.operators.Merger;
import de.uniwue.web.io.FileDatabase;
import de.uniwue.web.io.FilePathManager;
import de.uniwue.web.io.ImageLoader;
import de.uniwue.web.model.Book;
import de.uniwue.web.model.Point;
import de.uniwue.web.model.Polygon;
import de.uniwue.web.model.Rectangle;
import de.uniwue.web.model.Region;

/**
 * Segmenter using the LAREX project/algorithm
 *
 */
public class ImageProcessingFacade {

	public static Region merge(List<Region> segments) {
		ArrayList<RegionSegment> resultRegions = new ArrayList<RegionSegment>();
		for (Region segment : segments)
			resultRegions.add(segment.toRegionSegment());

		RegionSegment mergedRegion = Merger.lineMerge(resultRegions);
		MemoryCleaner.clean(resultRegions);

		Polygon coords = new Polygon();
		assert mergedRegion != null;
		for (org.opencv.core.Point regionPoint : mergedRegion.getPoints().toList()) {
			coords.addPoint(new Point(regionPoint.x, regionPoint.y));
		}
		MemoryCleaner.clean(mergedRegion);

		return new Region(mergedRegion.getId(), coords, mergedRegion.getType().toString());
	}

	public static Collection<List<Point>> extractContours(int pageNr, int bookID, FilePathManager fileManager, FileDatabase database) {
		Book book = database.getBook(bookID);
		File imagePath = fileManager.getImagePath(book.getPage(pageNr));

		Mat gray = ImageLoader.readGray(imagePath);
		Collection<MatOfPoint> contours = Contourextractor.fromGray(gray);
		MemoryCleaner.clean(gray);

		Collection<List<Point>> contourSegments = new ArrayList<>();
		for (final MatOfPoint contour : contours) {
			LinkedList<Point> points = new LinkedList<>();
			for (org.opencv.core.Point regionPoint : contour.toList()) {
				points.add(new Point(regionPoint.x, regionPoint.y));
			}
			contourSegments.add(points);
			MemoryCleaner.clean(contour);
		}

		return contourSegments;
	}

	/**
	 * Request to combine contours (point list) to a polygon of type paragraph.
	 *
	 * @param contours    Contours to combine
	 * @param accuracy    Accuracy of the combination process (between 0 and 100)
	 * @param fileManager Filemanager to load the book/page from
	 * @return Polygon that includes all contours
	 */
	public static Region combineContours(Collection<List<Point>> contours, int pageWidth, int pageHeight, int accuracy,
										 FilePathManager fileManager, FileDatabase database) {

		Collection<MatOfPoint> matContours = new ArrayList<>();
		for (List<Point> contour : contours) {
			org.opencv.core.Point[] matPoints = new org.opencv.core.Point[contour.size()];
			for (int index = 0; index < contour.size(); index++) {
				Point point = contour.get(index);
				matPoints[index] = new org.opencv.core.Point(point.getX(), point.getY());
			}
			matContours.add(new MatOfPoint(matPoints));
		}


		accuracy = Math.max(100,accuracy);

		double growth = 105 - 100/(accuracy/100.0);

		final MatOfPoint combined = Merger.smearMerge(matContours, new Size(pageWidth, pageHeight), growth, growth, 10);
		MemoryCleaner.clean(matContours);

		Polygon coords = new Polygon();
		for (org.opencv.core.Point regionPoint : combined.toList()) {
			coords.addPoint(new Point(regionPoint.x, regionPoint.y));
		}
		MemoryCleaner.clean(combined);
		return new Region(UUID.randomUUID().toString(), coords, RegionSubType.paragraph.toString());
	}

	/**
	 * Calculate a min area rectangle around a region segment
	 *
	 * @param segment
	 * @return
	 */
	public static Rectangle getMinAreaRectangle(Polygon segment) {
		final org.opencv.core.Point[] origPoints = segment.getPoints().stream()
				.map(p -> new org.opencv.core.Point(p.getX(),p.getY())).toArray(org.opencv.core.Point[]::new);

		final MatOfPoint2f origPointMap = new MatOfPoint2f(origPoints);
		final RotatedRect rotated = Imgproc.minAreaRect(origPointMap);

		MemoryCleaner.clean(origPointMap);

		final ArrayList<Point> points = new ArrayList<>();
		final org.opencv.core.Point center = rotated.center;
		final double angle = rotated.angle < -45 ? rotated.angle+90 : rotated.angle;
		final Size size = rotated.angle < -45 ? new Size(rotated.size.height,rotated.size.width) : rotated.size;

		final Function<Point, Point> asGlobal = (point) -> {
			final double x = Math.cos(angle) * point.getX() - Math.sin(angle) * point.getY() + center.x;
			final double y = Math.sin(angle) * point.getX() + Math.cos(angle) * point.getY() + center.y;
			return new Point(x,y);};

		points.add(asGlobal.apply(new Point(0,0)));
		points.add(asGlobal.apply(new Point(size.width,0)));
		points.add(asGlobal.apply(new Point(size.width,size.height)));
		points.add(asGlobal.apply(new Point(0,size.height)));

		return new Rectangle(points, size.height, size.width, angle, segment.isRelative());
	}

	/**
	 * Calculate a simplified polygon for each region in a given list of regions
	 * @param segments regions to be simplified
	 * @param tolerance simplifier tolerance
	 * @return list containing simplified regions
	 */
	public static List<Region> simplify(List<Region> segments, int tolerance) {
		ArrayList<Region> resultRegions = new ArrayList<Region>();
		for (Region segment : segments) {
			List<Point> coords = segment.getCoords().getPoints();
			List<org.locationtech.jts.geom.Coordinate> points = new ArrayList<>();

			//convert larex coords to jts coords
			for(Point coord : coords) {
				points.add(new Coordinate(coord.getX(),coord.getY()));
			}
			//add start coordinate to end of list to form linear ring
			points.add(new Coordinate(coords.get(0).getX(),coords.get(0).getY()));

			// simplify using jts libraries, preserving topology
			Coordinate[] coordinates = points.toArray(new Coordinate[points.size()]);
			//use lowest precision model to as we are using integer coordinates anyway
			GeometryFactory geometryFactory = new GeometryFactory(new PrecisionModel(PrecisionModel.FLOATING_SINGLE));
			LinearRing linearRing = new LinearRing(new CoordinateArraySequence(coordinates), geometryFactory);
			org.locationtech.jts.geom.Polygon polygon = new org.locationtech.jts.geom.Polygon(linearRing,null,geometryFactory);
			org.locationtech.jts.geom.Polygon simplifiedPolygon = (org.locationtech.jts.geom.Polygon) TopologyPreservingSimplifier.simplify(polygon,tolerance);

			//convert jts coords to larex coords
			ArrayList<Point> newCoords = new ArrayList<>();
			for(Coordinate coordinate : simplifiedPolygon.getCoordinates()) {
				newCoords.add(new Point(coordinate.getX(),coordinate.getY()));
			}
			//set simplified region coords
			segment.setCoords(new Polygon(newCoords));
			resultRegions.add(segment);
		}
		return resultRegions;
	}
}
